@@ -65,7 +65,7 @@ u_char pktversion = NTP_VERSION;
 /*
  * Don't jump if no set jmp.
  */
-VOLATILE int jump = 0;
+volatile int jump = 0;
 
 /*
  * Format values
