@@ -91,6 +91,7 @@ struct ifconf {
  * Block the interrupt, for critical sections.
  */
 #if defined(HAVE_SIGNALED_IO)
+static int sigio_block_count = 0;
 # define BLOCKIO()   ((void) block_sigio())
 # define UNBLOCKIO() ((void) unblock_sigio())
 #else
@@ -109,12 +110,12 @@ struct ifconf {
 /*
  * Memory allocation
  */
-u_long full_recvbufs;		/* number of recvbufs on fulllist */
-u_long free_recvbufs;		/* number of recvbufs on freelist */
+volatile u_long full_recvbufs;		/* number of recvbufs on fulllist */
+volatile u_long free_recvbufs;		/* number of recvbufs on freelist */
 
-static	struct recvbuf *freelist;	/* free buffers */
-static	struct recvbuf *fulllist;	/* lifo buffers with data */
-static	struct recvbuf *beginlist;	/* fifo buffers with data */
+volatile static	struct recvbuf *freelist;	/* free buffers */
+volatile static	struct recvbuf *fulllist;	/* lifo buffers with data */
+volatile static	struct recvbuf *beginlist;	/* fifo buffers with data */
 
 u_long total_recvbufs;		/* total recvbufs currently in use */
 u_long lowater_additions;	/* number of times we have added memory */
@@ -125,14 +126,14 @@ static	struct recvbuf initial_bufs[RECV_INIT];	/* initial allocation */
 /*
  * Other statistics of possible interest
  */
-u_long packets_dropped;	/* total number of packets dropped on reception */
-u_long packets_ignored;	/* packets received on wild card interface */
-u_long packets_received;	/* total number of packets received */
+volatile u_long packets_dropped;	/* total number of packets dropped on reception */
+volatile u_long packets_ignored;	/* packets received on wild card interface */
+volatile u_long packets_received;	/* total number of packets received */
 u_long packets_sent;	/* total number of packets sent */
 u_long packets_notsent;	/* total number of packets which couldn't be sent */
 
-u_long handler_calls;	/* number of calls to interrupt handler */
-u_long handler_pkts;	/* number of pkts received by handler */
+volatile u_long handler_calls;	/* number of calls to interrupt handler */
+volatile u_long handler_pkts;	/* number of pkts received by handler */
 u_long io_timereset;	/* time counters were reset */
 
 /*
@@ -171,6 +172,7 @@ extern int debug;
 static	int	create_sockets	P((u_int));
 static	int	open_socket	P((struct sockaddr_in *, int, int));
 static	void	close_socket	P((int));
+static	void	close_file		P((int));
 static	char *	fdbits		P((int, fd_set *));
 #ifdef HAVE_SIGNALED_IO
 static  int	init_clock_sig	P((struct refclockio *));
@@ -294,7 +296,7 @@ create_sockets(port)
 #if	_BSDI_VERSION >= 199510
   if (getifaddrs(&ifaddrs, &num_if) < 0)
     {
-      msyslog(LOG_ERR, "getifaddrs: %m");
+      msyslog(LOG_ERR, "create_sockets: getifaddrs() failed: %m");
       exit(1);
     }
 
@@ -399,7 +401,7 @@ create_sockets(port)
   if(ioctl(vs, I_STR, &ioc) < 0 ||
      ioc.ic_len < sizeof(struct ifreq))
     {
-      msyslog(LOG_ERR, "get interface configuration: %m");
+      msyslog(LOG_ERR, "create_sockets: ioctl(I_STR:SIOCGIFCONF) failed: %m - exiting");
       exit(1);
     }
 #  ifdef SIZE_RETURNED_IN_BUFFER
@@ -419,7 +421,7 @@ create_sockets(port)
   if (get_winnt_interfaces(&ifc) < 0)
 #  endif /* SYS_WINNT */
     {
-      msyslog(LOG_ERR, "get interface configuration: %m - exiting");
+      msyslog(LOG_ERR, "create_sockets: ioctl(SIOCGIFCONF) failed: %m - exiting");
       exit(1);
     }
 
@@ -451,13 +453,13 @@ create_sockets(port)
       ioc.ic_len = sizeof(struct ifreq);
       if(ioctl(vs, I_STR, &ioc))
 	{
-	  msyslog(LOG_ERR, "Failed to get interface flags: %m");
+	  msyslog(LOG_ERR, "create_sockets: ioctl(I_STR:SIOCGIFFLAGS) failed: %m");
 	  continue;
 	}
 #  else /* not STREAMS_TLI */
       if (ioctl(vs, SIOCGIFFLAGS, (char *)&ifreq) < 0)
 	{
-	  msyslog(LOG_ERR, "Failed to get interface flags: %m");
+	  msyslog(LOG_ERR, "create_sockets: ioctl(SIOCGIFFLAGS) failed: %m");
 	  continue;
 	}
 #  endif /* not STREAMS_TLI */
@@ -489,6 +491,7 @@ create_sockets(port)
 	}
 # endif /* not SUN_3_3_STINKS */
 
+#if 0
 # ifndef SYS_WINNT
 #  ifdef STREAMS_TLI
       ioc.ic_cmd = SIOCGIFADDR;
@@ -497,22 +500,22 @@ create_sockets(port)
       ioc.ic_len = sizeof(struct ifreq);
       if (ioctl(vs, I_STR, &ioc))
 	{
-	  msyslog(LOG_ERR, "Failed to get interface addr: %m");
+	  msyslog(LOG_ERR, "create_sockets: ioctl(I_STR:SIOCGIFADDR) failed: %m");
 	  continue;
 	}
 #  else /* not STREAMS_TLI */
       if (ioctl(vs, SIOCGIFADDR, (char *)&ifreq) < 0)
 	{
-	  msyslog(LOG_ERR, "Failed to get interface addr: %m");
+	  msyslog(LOG_ERR, "create_sockets: ioctl(SIOCGIFADDR) failed: %m");
 	  continue;
 	}
 #  endif /* not STREAMS_TLI */
-
 # endif /* not SYS_WINNT */
+#endif 0
 
       (void)strncpy(inter_list[i].name, ifreq.ifr_name,
 		    sizeof(inter_list[i].name));
-      inter_list[i].sin = *(struct sockaddr_in *)&ifreq.ifr_addr;
+      inter_list[i].sin = *(struct sockaddr_in *)&ifr->ifr_addr;
       inter_list[i].sin.sin_family = AF_INET;
       inter_list[i].sin.sin_port = port;
 
@@ -537,13 +540,13 @@ create_sockets(port)
 	  ioc.ic_len = sizeof(struct ifreq);
 	  if(ioctl(vs, I_STR, &ioc))
 	    {
-	      msyslog(LOG_ERR, "SIOCGIFBRDADDR fails");
+	      msyslog(LOG_ERR, "create_sockets: ioctl(I_STR:SIOCGIFBRDADDR) failed: %m");
 	      exit(1);
 	    }
 #  else /* not STREAMS_TLI */
 	  if (ioctl(vs, SIOCGIFBRDADDR, (char *)&ifreq) < 0)
 	    {
-	      msyslog(LOG_ERR, "SIOCGIFBRDADDR fails: %m");
+	      msyslog(LOG_ERR, "create_sockets: ioctl(SIOCGIFBRDADDR) failed: %m");
 	      exit(1);
 	    }
 #  endif /* not STREAMS_TLI */
@@ -566,13 +569,13 @@ create_sockets(port)
       ioc.ic_len = sizeof(struct ifreq);
       if(ioctl(vs, I_STR, &ioc))
 	{
-	  msyslog(LOG_ERR, "SIOCGIFNETMASK fails");
+	  msyslog(LOG_ERR, "create_sockets: ioctl(I_STR:SIOCGIFNETMASK) failed: %m");
 	  exit(1);
 	}
 #  else /* not STREAMS_TLI */
       if (ioctl(vs, SIOCGIFNETMASK, (char *)&ifreq) < 0)
 	{
-	  msyslog(LOG_ERR, "SIOCGIFNETMASK fails");
+	  msyslog(LOG_ERR, "create_sockets: ioctl(SIOCGIFNETMASK) failed: %m");
 	  exit(1);
 	}
 #  endif /* not STREAMS_TLI */
@@ -618,7 +621,7 @@ create_sockets(port)
       if (setsockopt(inter_list[i].fd, SOL_SOCKET, SO_REUSEADDR,
 		     (char *)&off, sizeof(off)))
 	{
-	  msyslog(LOG_ERR, "setsockopt SO_REUSEADDR off fails: %m");
+	  msyslog(LOG_ERR, "create_sockets: setsockopt(SO_REUSEADDR,off) failed: %m");
 	}
     }
 
@@ -684,7 +687,7 @@ io_setbclient()
 #ifdef	SYS_SOLARIS
       inter_list[i].bcast.sin_addr.s_addr = htonl(INADDR_ANY);
 #endif
-#if !defined(SYS_DOMAINOS) && !defined(SYS_LINUX)
+#ifdef OPEN_BCAST_SOCKET /* Was: !SYS_DOMAINOS && !SYS_LINUX */
       inter_list[i].bfd = open_socket(&inter_list[i].bcast, 0, 1);
       inter_list[i].flags |= INT_BCASTOPEN;
 #endif
@@ -831,7 +834,7 @@ io_multicast_del(addr)
 	continue;
       if (i != 0)
 	{
-	  /* we have an explicit fd, so we can slose it */
+	  /* we have an explicit fd, so we can close it */
 	  close_socket(inter_list[i].fd);
 	  memset((char *)&inter_list[i], 0, sizeof inter_list[0]);
 	  inter_list[i].fd = -1;
@@ -867,21 +870,18 @@ open_socket(addr, flags, turn_off_reuse)
   int on = 1, off = 0;
 
   /* create a datagram (UDP) socket */
-#ifndef WINNT
-  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
-    {
-      msyslog(LOG_ERR, "socket(AF_INET, SOCK_DGRAM, 0) failed: %m");
-      exit(1);
-      /*NOTREACHED*/
-    }
+  if (  (fd = socket(AF_INET, SOCK_DGRAM, 0)) 
+#ifndef SYS_WINNT
+		< 0
 #else
-  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET)
+		== INVALID_SOCKET
+#endif /* SYS_WINNT */
+	 )
     {
       msyslog(LOG_ERR, "socket(AF_INET, SOCK_DGRAM, 0) failed: %m");
       exit(1);
       /*NOTREACHED*/
     }
-#endif
 
   /* set SO_REUSEADDR since we will be binding the same port
      number on each interface */
@@ -925,50 +925,58 @@ open_socket(addr, flags, turn_off_reuse)
     maxactivefd = fd;
   FD_SET(fd, &activefds);
 
-#ifdef HAVE_SIGNALED_IO
-  init_socket_sig(fd);
-#else /* not HAVE_SIGNALED_IO */
-#endif /* not HAVE_SIGNALED_IO */
-
   /*
    * set non-blocking,
    */
 
-#if defined(O_NONBLOCK)
+#if defined(O_NONBLOCK) /* POSIX */
   if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0)
     {
       msyslog(LOG_ERR, "fcntl(O_NONBLOCK) fails: %m");
       exit(1);
       /*NOTREACHED*/
     }
-#else /* O_NONBLOCK */
-# if defined(FNDELAY)
+#elif defined(FNDELAY)
   if (fcntl(fd, F_SETFL, FNDELAY) < 0)
     {
       msyslog(LOG_ERR, "fcntl(FNDELAY) fails: %m");
       exit(1);
       /*NOTREACHED*/
     }
-# else /* FNDELAY */
-#  if defined(VMS) || defined(SYS_WINNT) /* VMS+UCX or SYS_WINNT */
-#   ifndef SYS_WINNT
+#elif defined(O_NDELAY) /* generally the same as FNDELAY */
+  if (fcntl(fd, F_SETFL, O_NDELAY) < 0)
+    {
+      msyslog(LOG_ERR, "fcntl(O_NDELAY) fails: %m");
+      exit(1);
+      /*NOTREACHED*/
+    }
+#elif defined(FIONBIO)
+# if defined(VMS)
   if (ioctl(fd,FIONBIO,&1) < 0)
-#   else
+# elif defined(SYS_WINNT)
   if (ioctlsocket(fd,FIONBIO,(u_long *) &on) == SOCKET_ERROR)
-#   endif /* SYS_WINNT */
+# else
+  if (ioctl(fd,FIONBIO,&on) < 0)
+# endif
     {
       msyslog(LOG_ERR, "ioctl(FIONBIO) fails: %m");
       exit(1);
+      /*NOTREACHED*/
     }
-#  else
+#elif defined(FIOSNBIO)
+  if (ioctl(fd,FIOSNBIO,&on) < 0)
+    {
+      msyslog(LOG_ERR, "ioctl(FIOSNBIO) fails: %m");
+      exit(1);
+      /*NOTREACHED*/
+    }
+#else
 #   include "Bletch: Need non blocking I/O!"
-#  endif /* VMS+UCX or SYS_WINNT */
-# endif /* FNDELAY */
-#endif /* O_NONBLOCK */
+#endif
 
 #ifdef HAVE_SIGNALED_IO
-#else /* not HAVE_SIGNALED_IO */
-#endif /* HAVE_SIGNALED_IO */
+  init_socket_sig(fd);
+#endif /* not HAVE_SIGNALED_IO */
 
   /*
    *  Turn off the SO_REUSEADDR socket option.  It apparently
@@ -1030,6 +1038,31 @@ close_socket(fd)
 }
 
 
+/*
+ * close_file - close a file and remove from the activefd list
+ * added 1/31/1997 Greg Schueman for Windows NT portability
+ */
+static void
+close_file(fd)
+     int fd;
+{
+  int i, newmax;
+
+  (void) close(fd);
+  FD_CLR( (u_int) fd, &activefds);
+
+  if (fd >= maxactivefd)
+    {
+      newmax = 0;
+      for (i = 0; i < maxactivefd; i++)
+	if (FD_ISSET(i, &activefds))
+	  newmax = i;
+      maxactivefd = newmax;
+    }
+}
+
+
+
 
 /*
  * findbcastinter - find broadcast interface corresponding to address
@@ -1164,6 +1197,7 @@ sendpkt(dest, inter, ttl, pkt, len)
 #ifdef SYS_WINNT
   DWORD err;
 #endif /* SYS_WINNT */
+
   /*
    * Send error cache. Empty slots have port == 0
    * Set ERRORCACHESIZE to 0 to disable
@@ -1221,7 +1255,7 @@ sendpkt(dest, inter, ttl, pkt, len)
       if (errno != EWOULDBLOCK && errno != ENOBUFS && slot < 0)
 #else
       err = WSAGetLastError();
-      if (err != WSAEWOULDBLOCK && err != WSAENOBUFS && slot <0)
+      if (err != WSAEWOULDBLOCK && err != WSAENOBUFS && slot < 0)
 #endif /* SYS_WINNT */
 	{
 	  /*
@@ -1245,7 +1279,10 @@ sendpkt(dest, inter, ttl, pkt, len)
        * He's not bad any more
        */
       if (slot >= 0)
-	badaddrs[slot].port = 0;
+	{
+	  msyslog(LOG_INFO, "Connection re-established to %s", ntoa(dest));
+	  badaddrs[slot].port = 0;
+	}
     }
 }
 
@@ -1273,6 +1310,7 @@ fdbits(count, set)
   return buffer;
 }
 
+
 /*
  * input_handler - receive packets asynchronously
  */
@@ -1286,10 +1324,18 @@ input_handler(cts)
   register int fd;
   struct timeval tvzero;
   int fromlen;
-  l_fp ts;
+  l_fp ts;			/* Timestamp at BOselect() gob */
+  l_fp ts_e;			/* Timestamp at EOselect() gob */
   fd_set fds;
+  int select_count = 0;
+#if 0
   int first = 1;
+#endif
+  static int handler_count = 0;
 
+  ++handler_count;
+  if (handler_count != 1)
+    msyslog(LOG_ERR, "input_handler: handler_count is %d!", handler_count);
   handler_calls++;
   ts = *cts;
 
@@ -1301,14 +1347,303 @@ input_handler(cts)
 
       fds = activefds;
       tvzero.tv_sec = tvzero.tv_usec = 0;
-      n = select(maxactivefd+1, &fds, (fd_set *)0, (fd_set *)0, &tvzero);
+      
+      /*
+       * If we have something to do, freeze a timestamp.
+       * See below for the other cases (nothing (left) to do or error)
+       */
+      while (0 < (n = select(maxactivefd+1, &fds, (fd_set *)0, (fd_set *)0, &tvzero)))
+	{
+#if 0
+	  if (!first) get_systime(&ts);
+	  first = 0;
+#endif
+	  ++select_count;
+#if 1
+	  ++handler_pkts;
+#else
+	  handler_pkts += n;
+#endif
+
+#ifdef REFCLOCK
+	  /*
+	   * Check out the reference clocks first, if any
+	   */
+	  if (refio != 0)
+	    {
+	      register struct refclockio *rp;
+
+	      for (rp = refio; rp != 0 && n > 0; rp = rp->next)
+		{
+		  fd = rp->fd;
+		  if (FD_ISSET(fd, &fds))
+		    {
+		      n--;
+		      if (free_recvbufs == 0)
+			{
+			  char buf[RX_BUFF_SIZE];
+
+#ifndef SYS_WINNT
+			  (void) read(fd, buf, sizeof buf);
+#else
+			  (void) ReadFile((HANDLE)fd, buf, (DWORD)sizeof buf, NULL, NULL);
+#endif /* SYS_WINNT */
+			  packets_dropped++;
+#if 1
+			  goto select_again;
+#else
+			  continue;
+#endif
+			}
+
+		      rb = freelist;
+		      freelist = rb->next;
+		      free_recvbufs--;
+
+		      i = (rp->datalen == 0
+			   || rp->datalen > sizeof(rb->recv_space))
+			? sizeof(rb->recv_space) : rp->datalen;
+#ifndef SYS_WINNT				
+		      rb->recv_length =
+			read(fd, (char *)&rb->recv_space, i)
+#else  /* SYS_WINNT */
+			ReadFile((HANDLE)fd, (char *)&rb->recv_space, (DWORD)i,
+				 (LPDWORD)&(rb->recv_length), NULL)
+#endif /* SYS_WINNT */
+			;
+
+		      if (rb->recv_length == -1)
+			{
+			  msyslog(LOG_ERR, "clock read fd %d: %m", fd);
+			  rb->next = freelist;
+			  freelist = rb;
+			  free_recvbufs++;
+#if 1
+			  goto select_again;
+#else
+			  continue;
+#endif
+			}
+	
+		      /*
+		       * Got one.  Mark how and when it got here,
+		       * put it on the full list and do bookkeeping.
+		       */
+		      rb->recv_srcclock = rp->srcclock;
+		      rb->dstadr = 0;
+		      rb->fd = fd;
+		      rb->recv_time = ts;
+		      rb->receiver = rp->clock_recv;
+
+		      if (fulllist == 0)
+			{
+			  beginlist = rb;
+			  rb->next = 0;
+			}
+		      else
+			{
+			  rb->next = fulllist->next;
+			  fulllist->next = rb;
+			}
+		      fulllist = rb;
+		      full_recvbufs++;
+
+		      rp->recvcount++;
+		      packets_received++;
+		    }
+		}
+	    }
+#endif /* REFCLOCK */
+
+	  /*
+	   * Loop through the interfaces looking for data to read.
+	   */
+	  for (i = ninterfaces - 1; (i >= 0) && (n > 0); i--)
+	    {
+	      for (doing = 0; (doing < 2) && (n > 0); doing++)
+		{
+		  if (doing == 0)
+		    {
+		      fd = inter_list[i].fd;
+		    }
+		  else
+		    {
+		      if (!(inter_list[i].flags & INT_BCASTOPEN))
+			break;
+		      fd = inter_list[i].bfd;
+		    }
+		  if (fd < 0) continue;
+		  if (FD_ISSET(fd, &fds))
+		    {
+		      n--;
+
+		      /*
+		       * Get a buffer and read the frame.  If we
+		       * haven't got a buffer, or this is received
+		       * on the wild card socket, just dump the
+		       * packet.
+		       */
+		      if (
+#ifndef SYS_WINNT
+			  (!(free_recvbufs && (i == 0) &&
+			     (inter_list[i].flags & INT_MULTICAST)))
+#else
+			  /* For Win/NT we are creating just one socket,
+			   * i is 0, and INT_MULTICAST flag is not set.
+			   */
+			  (!free_recvbufs)
+#endif /* SYS_WINNT */
+			  )
+			{
+			  if (
+#ifdef UDP_WILDCARD_DELIVERY
+			      /*
+			       * these guys manage to put properly addressed
+			       * packets into the wildcard queue
+			       */
+			      (free_recvbufs == 0)
+#else
+			      ((i == 0) || (free_recvbufs == 0))
+#endif
+			      )
+			    {
+			      char buf[RX_BUFF_SIZE];
+			      struct sockaddr from;
+
+			      fromlen = sizeof from;
+			      (void) recvfrom(fd, buf,
+					      sizeof(buf), 0,
+					      &from, &fromlen);
+#ifdef DEBUG
+			      if (debug)
+				printf("%s on %d(%lu) fd=%d from %s\n",
+				       (i) ? "drop" : "ignore",
+				       i, free_recvbufs, fd,
+				       inet_ntoa(((struct sockaddr_in *) &from)->sin_addr));
+#endif
+			      if (i == 0)
+				packets_ignored++;
+			      else
+				packets_dropped++;
+#if 1
+			      goto select_again;
+#else
+			      continue;
+#endif
+			    }
+			}
+
+		      rb = freelist;
+
+		      fromlen = sizeof(struct sockaddr_in);
+		      rb->recv_length = recvfrom(fd,
+						 (char *)&rb->recv_space,
+						 sizeof(rb->recv_space), 0,
+						 (struct sockaddr *)&rb->recv_srcadr,
+						 &fromlen);
+		      if (rb->recv_length > 0)
+		        {
+		          freelist = rb->next;
+		          free_recvbufs--;
+		        }
+		      else if (rb->recv_length == 0
+#ifdef EWOULDBLOCK
+			       || errno==EWOULDBLOCK
+#endif
+#ifdef EAGAIN
+			       || errno==EAGAIN
+#endif
+		              )
+		        continue;
+		      else
+		        {
+			  msyslog(LOG_ERR, "recvfrom() fd=%d: %m", fd);
+#ifdef DEBUG
+			  if (debug)
+			    printf("input_handler: fd=%d dropped (bad recvfrom)\n", fd);
+#endif
+			  continue;
+			}
+#ifdef DEBUG
+		      if (debug)
+			printf("input_handler: fd=%d length %d from %08lx %s\n",
+			       fd, rb->recv_length,
+			       (u_long)ntohl(rb->recv_srcadr.sin_addr.s_addr) &
+			       0x00000000ffffffff,
+			       inet_ntoa(rb->recv_srcadr.sin_addr));
+#endif
+		      
+		      /*
+		       * Got one.  Mark how and when it got here,
+		       * put it on the full list and do bookkeeping.
+		       */
+		      rb->dstadr = &inter_list[i];
+		      rb->fd = fd;
+		      rb->recv_time = ts;
+		      rb->receiver = receive;
+
+		      if (fulllist == 0)
+			{
+			  beginlist = rb;
+			  rb->next = 0;
+			}
+		      else
+			{
+			  rb->next = fulllist->next;
+			  fulllist->next = rb;
+			}
+		      fulllist = rb;
+		      full_recvbufs++;
+	
+		      inter_list[i].received++;
+		      packets_received++;
+		      goto select_again;
+		    }
+		  /* Check more interfaces */
+		}
+	    }
+	select_again:;
+	  /*
+	   * Done everything from that select.  Poll again.
+	   */
+	}
 
       /*
-       * If nothing to do, just return.  If an error occurred, complain
-       * and return.  If we've got some, freeze a timestamp.
+       * If nothing more to do, try again.
+       * If nothing to do, just return.
+       * If an error occurred, complain and return.
        */
       if (n == 0)
-	return;
+	{
+	  if (select_count == 0) /* We really had nothing to do */
+	    {
+	      if (debug)
+		msyslog(LOG_DEBUG, "input_handler: select() returned 0");
+	      --handler_count;
+	      return;
+	    }
+	  /* We've done our work */
+	  get_systime(&ts_e);
+	  /*
+	   * (ts_e - ts) is the amount of time we spent processing
+	   * this gob of file descriptors.  Log it.
+	   */
+	  L_SUB(&ts_e, &ts);
+	  if (debug > 3)
+	    msyslog(LOG_INFO, "input_handler: Processed a gob of fd's in %s msec", lfptoms(&ts_e, 6));
+
+#if 0
+	  /*
+	   * We'll re-start the for(;;) loop now.
+	   * Use the ending timestamp as the received timestamp
+	   */
+	  ts = ts_e;
+#else
+	  /* No, for now just bail. */
+	  --handler_count;
+	  return;
+#endif
+	}
       else if (n == -1)
 	{
 #ifndef SYS_WINNT
@@ -1343,243 +1678,15 @@ input_handler(cts)
 		    )
 		  msyslog(LOG_ERR, "Bad file descriptor %d", i);
 	    }
+	  --handler_count;
 	  return;
 	}
-      if (!first) get_systime(&ts);
-      first = 0;
-      handler_pkts += n;
-
-#ifdef REFCLOCK
-      /*
-       * Check out the reference clocks first, if any
-       */
-      if (refio != 0)
-	{
-	  register struct refclockio *rp;
-
-	  for (rp = refio; rp != 0 && n > 0; rp = rp->next)
-	    {
-	      fd = rp->fd;
-	      if (FD_ISSET(fd, &fds))
-		{
-		  n--;
-		  FD_CLR(fd, &fds); /* HMS: Added 961211 */
-		  if (free_recvbufs == 0)
-		    {
-		      char buf[RX_BUFF_SIZE];
-
-#if 0	/* GMS */
-		      (void) recv(fd, buf, sizeof buf, 0);
-#else
-# ifndef SYS_WINNT
-		      (void) read(fd, buf, sizeof buf);
-# else
-		      (void) ReadFile((HANDLE)fd, buf, (DWORD)sizeof buf, NULL, NULL);
-# endif /* SYS_WINNT */
-#endif
-		      packets_dropped++;
-		      continue;
-		    }
-
-		  rb = freelist;
-		  freelist = rb->next;
-		  free_recvbufs--;
-
-		  i = (rp->datalen == 0
-		       || rp->datalen > sizeof(rb->recv_space))
-		    ? sizeof(rb->recv_space) : rp->datalen;
-#if 0	/* GMS */
-		  rb->recv_length =
-		    recv(fd, (char *)&rb->recv_space, i, 0);
-#else
-# ifndef SYS_WINNT				
-		  rb->recv_length =
-		    read(fd, (char *)&rb->recv_space, i);
-# else
-		  ReadFile((HANDLE)fd, (char *)&rb->recv_space, (DWORD)i,
-			   (LPDWORD)&(rb->recv_length), NULL);
-# endif /* SYS_WINNT */
-#endif
-
-		  if (rb->recv_length == -1)
-		    {
-		      msyslog(LOG_ERR, "clock read fd %d: %m",	fd);
-		      rb->next = freelist;
-		      freelist = rb;
-		      free_recvbufs++;
-		      continue;
-		    }
-	
-		  /*
-		   * Got one.  Mark how and when it got here,
-		   * put it on the full list and do bookkeeping.
-		   */
-		  rb->recv_srcclock = rp->srcclock;
-		  rb->dstadr = 0;
-		  rb->fd = fd;
-		  rb->recv_time = ts;
-		  rb->receiver = rp->clock_recv;
-
-		  if (fulllist == 0)
-		    {
-		      beginlist = rb;
-		      rb->next = 0;
-		    }
-		  else
-		    {
-		      rb->next = fulllist->next;
-		      fulllist->next = rb;
-		    }
-		  fulllist = rb;
-		  full_recvbufs++;
-
-		  rp->recvcount++;
-		  packets_received++;
-		}
-	    }
-	}
-#endif
-
-      /*
-       * Loop through the interfaces looking for data to read.
-       */
-      for (i = ninterfaces - 1; (i >= 0) && (n > 0); i--)
-	{
-	  for (doing = 0; (doing < 2) && (n > 0); doing++)
-	    {
-	      if (doing == 0)
-		{
-		  fd = inter_list[i].fd;
-		}
-	      else
-		{
-		  if (!(inter_list[i].flags & INT_BCASTOPEN))
-		    break;
-		  fd = inter_list[i].bfd;
-		}
-	      if (fd < 0) continue;
-	      if (FD_ISSET(fd, &fds))
-		{
-		  n--;
-		  FD_CLR(fd, &fds); /* HMS: Added 961211 */
-
-		  /*
-		   * Get a buffer and read the frame.  If we
-		   * haven't got a buffer, or this is received
-		   * on the wild card socket, just dump the
-		   * packet.
-		   */
-		  if (
-#ifndef SYS_WINNT
-		      (!(free_recvbufs && (i == 0) &&
-			 (inter_list[i].flags & INT_MULTICAST)))
-#else
-		      /* For Win/NT we are creating just one socket,
-		       * i is 0, and INT_MULTICAST flag is not set.
-		       */
-		      (!free_recvbufs)
-#endif /* SYS_WINNT */
-		      )
-		    {
-		      if (
-#ifdef UDP_WILDCARD_DELIVERY
-			  /*
-			   * these guys manage to put properly addressed
-			   * packets into the wildcard queue
-			   */
-			  (free_recvbufs == 0)
-#else
-			  ((i == 0) || (free_recvbufs == 0))
-#endif
-			  )
-			{
-			  char buf[RX_BUFF_SIZE];
-			  struct sockaddr from;
-
-			  fromlen = sizeof from;
-			  (void) recvfrom(fd, buf,
-					  sizeof(buf), 0,
-					  &from, &fromlen);
-#ifdef DEBUG
-			  if (debug)
-			    printf("%s on %d(%lu) fd=%d from %s\n",
-				   (i) ? "drop" : "ignore",
-				   i, free_recvbufs, fd,
-				   inet_ntoa(((struct sockaddr_in *) &from)->sin_addr));
-#endif
-			  if (i == 0)
-			    packets_ignored++;
-			  else
-			    packets_dropped++;
-			  continue;
-			}
-		    }
-
-		  rb = freelist;
-		  freelist = rb->next;
-		  free_recvbufs--;
-
-		  fromlen = sizeof(struct sockaddr_in);
-		  rb->recv_length = recvfrom(fd,
-					     (char *)&rb->recv_space,
-					     sizeof(rb->recv_space), 0,
-					     (struct sockaddr *)&rb->recv_srcadr,
-					     &fromlen);
-		  if (rb->recv_length == -1)
-		    {
-		      msyslog(LOG_ERR, "recvfrom() fd=%d: %m", fd);
-		      rb->next = freelist;
-		      freelist = rb;
-		      free_recvbufs++;
-#ifdef DEBUG
-		      if (debug)
-			printf("input_handler: fd=%d dropped (bad recvfrom)\n", fd);
-#endif
-		      continue;
-		    }
-#ifdef DEBUG
-		  if (debug)
-		    printf("input_handler: fd=%d length %d from %08lx %s\n",
-			   fd, rb->recv_length,
-			   (u_long)ntohl(rb->recv_srcadr.sin_addr.s_addr) &
-			   0x00000000ffffffff,
-			   inet_ntoa(rb->recv_srcadr.sin_addr));
-#endif
-
-		  /*
-		   * Got one.  Mark how and when it got here,
-		   * put it on the full list and do bookkeeping.
-		   */
-		  rb->dstadr = &inter_list[i];
-		  rb->fd = fd;
-		  rb->recv_time = ts;
-		  rb->receiver = receive;
-
-		  if (fulllist == 0)
-		    {
-		      beginlist = rb;
-		      rb->next = 0;
-		    }
-		  else
-		    {
-		      rb->next = fulllist->next;
-		      fulllist->next = rb;
-		    }
-		  fulllist = rb;
-		  full_recvbufs++;
-	
-		  inter_list[i].received++;
-		  packets_received++;
-		}
-	    }
-	}
-      /*
-       * Done everything from that select.  Poll again.
-       */
     }
+  msyslog(LOG_ERR, "input_handler: fell out of infinite for(;;) loop!");
+  --handler_count;
+  return;
 }
 
-
 /*
  * findinterface - utility used by other modules to find an interface
  *		   given an address.
@@ -1717,10 +1824,9 @@ io_closeclock(rio)
     }
   
   /*
-   * Close the descriptor.  close_socket does the right thing despite
-   * the misnomer.
+   * Close the descriptor.
    */
-  close_socket(rio->fd);
+  close_file(rio->fd);
 }
 #endif	/* REFCLOCK */
 
@@ -1894,107 +2000,72 @@ init_socket_sig(fd)
       exit(1);
     }
 }
-# else
-/*
- * Socket SIGPOLL initialization routines.
- * Special cases first!
- */
-/* Was: defined(SYS_HPUX) || defined(SYS_LINUX) */
-#  if defined(FIOSNBIO) && defined(FIOASYNC)
-#define SOCKET_DONE
+# else /* USE_UDP_SIGPOLL */
 {
-  int pgrp, on = 1;
-
-  /*
-   * Big difference here for HP-UX ... why can't life be easy ?
-   */
-  if (ioctl(fd, FIOSNBIO, (char *)&on) == -1)
-    {
-      msyslog(LOG_ERR, "ioctl(FIOSNBIO) fails: %m");
-      exit(1);
-      /*NOTREACHED*/
-    }
+  int pgrp;
+# ifdef FIOASYNC
+  int on = 1;
+# endif
 
+#  if defined(FIOASYNC)
   if (ioctl(fd, FIOASYNC, (char *)&on) == -1)
     {
       msyslog(LOG_ERR, "ioctl(FIOASYNC) fails: %m");
       exit(1);
       /*NOTREACHED*/
     }
+#  elif defined(FASYNC)
+  {
+    int flags;
 
-#   ifdef UDP_BACKWARDS_SETOWN
+    if ((flags = fcntl(fd, F_GETFL, 0)) == -1)
+      {
+        msyslog(LOG_ERR, "fcntl(F_GETFL) fails: %m");
+        exit(1);
+        /*NOTREACHED*/
+      }
+    if (fcntl(fd, F_SETFL, flags|FASYNC) < 0)
+      {
+        msyslog(LOG_ERR, "fcntl(...|FASYNC) fails: %m");
+        exit(1);
+        /*NOTREACHED*/
+      }
+   }
+#  else
+#   include "Bletch: Need asynchronous I/O!"
+#  endif
+
+#  ifdef UDP_BACKWARDS_SETOWN
   pgrp = -getpid();
-#   else
+#  else
   pgrp = getpid();
-#   endif
+#  endif
+
+#  if defined(SIOCSPGRP)
   if (ioctl(fd, SIOCSPGRP, (char *)&pgrp) == -1)
     {
       msyslog(LOG_ERR, "ioctl(SIOCSPGRP) fails: %m");
       exit(1);
       /*NOTREACHED*/
     }
-}
-#  endif /* SYS_HPUX || SYS_LINUX: FIOSNBIO && FIOASYNC */
-/* Was:  defined(SYS_AIX) && !defined(_BSD) */
-#  if !defined(_BSD) && defined(_AIX) && defined(FIOASYNC) && defined(FIOSETOWN)
-/*
- * SYSV compatibility mod under AIX
- */
-#define SOCKET_DONE
-{
-  int pgrp, on = 1;
-
-  if (ioctl(fd, FIOASYNC, (char *)&on) == -1)
-    {
-      msyslog(LOG_ERR, "ioctl(FIOASYNC) fails: %m");
-      exit(1);
-      /*NOTREACHED*/
-    }
-  pgrp = -getpid();
+#  elif defined(FIOSETOWN)
   if (ioctl(fd, FIOSETOWN, (char*)&pgrp) == -1)
     {
       msyslog(LOG_ERR, "ioctl(FIOSETOWN) fails: %m");
       exit(1);
       /*NOTREACHED*/
     }
-
-  if (fcntl(fd, F_SETFL, FNDELAY|FASYNC) < 0)
-    {
-      msyslog(LOG_ERR, "fcntl(FNDELAY|FASYNC) fails: %m");
-      exit(1);
-      /*NOTREACHED*/
-    }
-}
-#  endif /* AIX && !BSD: !_BSD && FIOASYNC && FIOSETOWN */
-#  ifndef SOCKET_DONE
-{
-  int pid = getpid();
-
-#   if defined(UDP_BACKWARDS_SETOWN)
-  /*
-   * The way Sun did it as recently as SunOS 3.5.  Only
-   * in the case of sockets, of course, just to confuse
-   * the issue.  Don't they even bother to test the stuff
-   * they send out?  Ibid for Ultrix 2.0
-   */
-  pid = -pid;
-#   endif /* UDP_BACKWARDS_SETOWN */
-  if (fcntl(fd, F_SETOWN, pid) == -1)
+#  elif defined(F_SETOWN)
+  if (fcntl(fd, F_SETOWN, pgrp) == -1)
     {
       msyslog(LOG_ERR, "fcntl(F_SETOWN) fails: %m");
       exit(1);
-    }
-  /*
-   * set non-blocking, async I/O on the descriptor
-   */
-  if (fcntl(fd, F_SETFL, FNDELAY|FASYNC) < 0)
-    {
-      msyslog(LOG_ERR, "fcntl(FNDELAY|FASYNC) fails: %m");
-      exit(1);
       /*NOTREACHED*/
     }
+#  else
+#   include "Bletch: Need to set process(group) to receive SIG(IO|POLL)"
+#  endif
 }
-#  endif /* SOCKET_DONE */
 # endif /* USE_UDP_SIGPOLL */
 
 
@@ -2002,23 +2073,12 @@ static RETSIGTYPE
 sigio_handler(sig)
      int sig;
 {
+  int saved_errno = errno;
   l_fp ts;
 
-# ifdef HAVE_SIGSET
-  /*
-   * This should not be necessary for a signal previously set with
-   * sigset().
-   */
-#  if defined(USE_SIGIO)
-  (void) sigset(SIGIO, sigio_handler);
-#  endif
-#  if defined(USE_SIGPOLL)
-  (void) sigset(SIGPOLL, sigio_handler);
-#  endif
-# endif /* HAVE_SIGSET */
-
   get_systime(&ts);
   (void)input_handler(&ts);
+  errno = saved_errno;
 }
 
 /*
@@ -2028,51 +2088,12 @@ sigio_handler(sig)
 static void
 set_signal()
 {
-  int n;
-  struct sigaction vec;
-
-  sigemptyset(&vec.sa_mask);
-
 #  ifdef USE_SIGIO
-  sigaddset(&vec.sa_mask, SIGIO);
-#  endif
+  (void) signal_no_reset(SIGIO, sigio_handler);
+# endif
 #  ifdef USE_SIGPOLL
-  sigaddset(&vec.sa_mask, SIGPOLL);
-#  endif
-  vec.sa_flags = 0;
-
-#  if defined(USE_SIGIO)
-  vec.sa_handler =  sigio_handler;
-
-  while (1)
-    {
-      n = sigaction(SIGIO, &vec, NULL);
-      if (n == -1 && errno == EINTR) continue;
-      break;
-    }
-
-  if (n == -1)
-    {
-      perror("sigaction");
-      exit(1);
-    }
-#  endif
-#  if defined(USE_SIGPOLL)
-  vec.sa_handler =  sigio_handler;
-
-  while (1)
-    {
-      n = sigaction(SIGPOLL, &vec, NULL);
-      if (n == -1 && errno == EINTR) continue;
-      break;
-    }
-
-  if (n == -1)
-    {
-      perror("sigaction");
-      exit(1);
-    }
-#  endif
+  (void) signal_no_reset(SIGPOLL, sigio_handler);
+# endif
 }
 
 void
@@ -2080,15 +2101,21 @@ block_io_and_alarm()
 {
   sigset_t set;
 
-  sigemptyset(&set);
+  if (sigemptyset(&set))
+    msyslog(LOG_ERR, "block_io_and_alarm: sigemptyset() failed: %m");
 #  if defined(USE_SIGIO)
-  sigaddset(&set, SIGIO);
+  if (sigaddset(&set, SIGIO))
+    msyslog(LOG_ERR, "block_io_and_alarm: sigaddset(SIGIO) failed: %m");
 #  endif
 #  if defined(USE_SIGPOLL)
-  sigaddset(&set, SIGPOLL);
+  if (sigaddset(&set, SIGPOLL))
+    msyslog(LOG_ERR, "block_io_and_alarm: sigaddset(SIGPOLL) failed: %m");
 #  endif
-  sigaddset(&set, SIGALRM);
-  sigprocmask(SIG_BLOCK, &set, NULL);
+  if (sigaddset(&set, SIGALRM))
+    msyslog(LOG_ERR, "block_io_and_alarm: sigaddset(SIGALRM) failed: %m");
+
+  if (sigprocmask(SIG_BLOCK, &set, NULL))
+    msyslog(LOG_ERR, "block_io_and_alarm: sigprocmask() failed: %m");
 }
 
 static void
@@ -2096,14 +2123,25 @@ block_sigio()
 {
   sigset_t set;
 
-  sigemptyset(&set);
+  ++sigio_block_count;
+  if (sigio_block_count > 1)
+    msyslog(LOG_INFO, "block_sigio: sigio_block_count > 1");
+  if (sigio_block_count < 1)
+    msyslog(LOG_INFO, "block_sigio: sigio_block_count < 1");
+
+  if (sigemptyset(&set))
+    msyslog(LOG_ERR, "block_sigio: sigemptyset() failed: %m");
 #  if defined(USE_SIGIO)
-  sigaddset(&set, SIGIO);
+  if (sigaddset(&set, SIGIO))
+    msyslog(LOG_ERR, "block_sigio: sigaddset(SIGIO) failed: %m");
 #  endif
 #  if defined(USE_SIGPOLL)
-  sigaddset(&set, SIGPOLL);
+  if (sigaddset(&set, SIGPOLL))
+    msyslog(LOG_ERR, "block_sigio: sigaddset(SIGPOLL) failed: %m");
 #  endif
-  sigprocmask(SIG_BLOCK, &set, NULL);
+
+  if (sigprocmask(SIG_BLOCK, &set, NULL))
+    msyslog(LOG_ERR, "block_sigio: sigprocmask() failed: %m");
 }
 
 void
@@ -2111,16 +2149,22 @@ unblock_io_and_alarm()
 {
   sigset_t unset;
 
-  sigemptyset(&unset);
+  if (sigemptyset(&unset))
+    msyslog(LOG_ERR, "unblock_io_and_alarm: sigemptyset() failed: %m");
 
 #  if defined(USE_SIGIO)
-  sigaddset(&unset, SIGIO);
+  if (sigaddset(&unset, SIGIO))
+    msyslog(LOG_ERR, "unblock_io_and_alarm: sigaddset(SIGIO) failed: %m");
 #  endif
 #  if defined(USE_SIGPOLL)
-  sigaddset(&unset, SIGPOLL);
+  if (sigaddset(&unset, SIGPOLL))
+    msyslog(LOG_ERR, "unblock_io_and_alarm: sigaddset(SIGPOLL) failed: %m");
 #  endif
-  sigaddset(&unset, SIGALRM);
-  sigprocmask(SIG_UNBLOCK, &unset, NULL);
+  if (sigaddset(&unset, SIGALRM))
+    msyslog(LOG_ERR, "unblock_io_and_alarm: sigaddset(SIGALRM) failed: %m");
+
+  if (sigprocmask(SIG_UNBLOCK, &unset, NULL))
+    msyslog(LOG_ERR, "unblock_io_and_alarm: sigprocmask() failed: %m");
 }
 
 static
@@ -2129,15 +2173,26 @@ unblock_sigio()
 {
   sigset_t unset;
 
-  sigemptyset(&unset);
+  --sigio_block_count;
+  if (sigio_block_count > 0)
+    msyslog(LOG_INFO, "unblock_sigio: sigio_block_count > 0");
+  if (sigio_block_count < 0)
+    msyslog(LOG_INFO, "unblock_sigio: sigio_block_count < 0");
+
+  if (sigemptyset(&unset))
+    msyslog(LOG_ERR, "unblock_sigio: sigemptyset() failed: %m");
 
 #  if defined(USE_SIGIO)
-  sigaddset(&unset, SIGIO);
+  if (sigaddset(&unset, SIGIO))
+    msyslog(LOG_ERR, "unblock_sigio: sigaddset(SIGIO) failed: %m");
 #  endif
 #  if defined(USE_SIGPOLL)
-  sigaddset(&unset, SIGPOLL);
+  if (sigaddset(&unset, SIGPOLL))
+    msyslog(LOG_ERR, "unblock_sigio: sigaddset(SIGPOLL) failed: %m");
 #  endif
-  sigprocmask(SIG_UNBLOCK, &unset, NULL);
+
+  if (sigprocmask(SIG_UNBLOCK, &unset, NULL))
+    msyslog(LOG_ERR, "unblock_sigio: sigprocmask() failed: %m");
 }
 
 void
@@ -2145,17 +2200,22 @@ wait_for_signal()
 {
   sigset_t old;
   
-  sigprocmask(SIG_UNBLOCK, NULL, &old);
+  if (sigprocmask(SIG_UNBLOCK, NULL, &old))
+    msyslog(LOG_ERR, "wait_for_signal: sigprocmask() failed: %m");
   
 #  if defined(USE_SIGIO)
-  sigdelset(&old, SIGIO);
+  if (sigdelset(&old, SIGIO))
+    msyslog(LOG_ERR, "wait_for_signal: sigdelset(SIGIO) failed: %m");
 #  endif
 #  if defined(USE_SIGPOLL)
-  sigdelset(&old, SIGPOLL);
+  if (sigdelset(&old, SIGPOLL))
+    msyslog(LOG_ERR, "wait_for_signal: sigdelset(SIGPOLL) failed: %m");
 #  endif
-  sigdelset(&old, SIGALRM);
+  if (sigdelset(&old, SIGALRM))
+    msyslog(LOG_ERR, "wait_for_signal: sigdelset(SIGALRM) failed: %m");
   
-  sigsuspend(&old);
+  if (sigsuspend(&old) && (errno != EINTR))
+    msyslog(LOG_ERR, "wait_for_signal: sigsuspend() failed: %m");
 }
 
 # else /* !HAVE_SIGACTION */
@@ -2170,16 +2230,24 @@ block_io_and_alarm()
   int mask;
   
   mask = sigmask(SIGIO) | sigmask(SIGALRM);
-  (void)sigblock(mask);
+  if (sigblock(mask))
+    msyslog(LOG_ERR, "block_io_and_alarm: sigblock() failed: %m");
 }
 
 static void
 block_sigio()
 {
   int mask;
-  
+
+  ++sigio_block_count;
+  if (sigio_block_count > 1)
+    msyslog(LOG_INFO, "block_sigio: sigio_block_count > 1");
+  if (sigio_block_count < 1)
+    msyslog(LOG_INFO, "block_sigio: sigio_block_count < 1");
+
   mask = sigmask(SIGIO);
-  (void)sigblock(mask);
+  if (sigblock(mask))
+    msyslog(LOG_ERR, "block_sigio: sigblock() failed: %m");
 }
 
 static void
@@ -2196,18 +2264,23 @@ unblock_io_and_alarm()
   mask = sigmask(SIGIO) | sigmask(SIGALRM);
   omask = sigblock(0);
   omask &= ~mask;
-  (void)sigsetmask(omask);
+  (void) sigsetmask(omask);
 }
 
 static void
 unblock_sigio()
 {
   int mask, omask;
-  
+
+  --sigio_block_count;
+  if (sigio_block_count > 0)
+    msyslog(LOG_INFO, "unblock_sigio: sigio_block_count > 0");
+  if (sigio_block_count < 0)
+    msyslog(LOG_INFO, "unblock_sigio: sigio_block_count < 0");
   mask = sigmask(SIGIO);
   omask = sigblock(0);
   omask &= ~mask;
-  (void)sigsetmask(omask);
+  (void) sigsetmask(omask);
 }
 
 void
@@ -2218,7 +2291,8 @@ wait_for_signal()
   mask = sigmask(SIGIO) | sigmask(SIGALRM);
   omask = sigblock(0);
   omask &= ~mask;
-  sigpause(omask);
+  if (sigpause(omask) && (errno != EINTR))
+    msyslog(LOG_ERR, "wait_for_signal: sigspause() failed: %m");
 }
 # endif /* HAVE_SIGACTION */
 #endif /* HAVE_SIGNALED_IO */
