@@ -776,6 +776,17 @@ true_doevent(peer, event)
 				abort();
 			}
 			break;
+			/*
+			 * An expedient hack to prevent lint complaints,
+			 * these don't actually need to be used here...
+			 */
+		case s_Init:
+		case s_F18:
+		case s_F50:
+		case s_Start:
+		case s_Auto:
+		case s_Max:
+		  msyslog(LOG_INFO, "TRUE: state %s is unexpected!", stateStr(up->state));
 		}
 		break;
 	default:
